@@ -5,33 +5,38 @@ Common utility functions, mostly related to loading configuration data.
 import os
 from os import getenv
 from os.path import abspath, dirname
+from sys import exit
 from typing import Dict, Optional
 
 from toml import TomlDecodeError
 
 from yd_commands.args import ARGS_PARSER
 from yd_commands.config_types import (
-    CR_BATCH_SIZE_DEFAULT,
-    DEFAULT_URL,
-    TASK_BATCH_SIZE_DEFAULT,
-    YD_KEY,
-    YD_NAMESPACE,
-    YD_SECRET,
-    YD_TAG,
-    YD_URL,
     ConfigCommon,
     ConfigWorkerPool,
     ConfigWorkRequirement,
 )
 from yd_commands.printing import print_error, print_log
 from yd_commands.property_names import *
+from yd_commands.settings import (
+    CR_BATCH_SIZE_DEFAULT,
+    DEFAULT_URL,
+    TASK_BATCH_SIZE_DEFAULT,
+    TOML_VAR_NESTED_DEPTH,
+    YD_KEY,
+    YD_NAMESPACE,
+    YD_SECRET,
+    YD_TAG,
+    YD_URL,
+)
 from yd_commands.type_check import check_list, check_str
 from yd_commands.utils import pathname_relative_to_config_file
 from yd_commands.validate_properties import validate_properties
 from yd_commands.variables import (
     add_substitutions,
     load_toml_file_with_variable_substitutions,
-    substitute_variable_str,
+    process_variable_substitutions,
+    process_variable_substitutions_in_dict_insitu,
 )
 
 # CLI > YD_CONF > 'config.toml'
@@ -79,7 +84,7 @@ def load_config_common() -> ConfigCommon:
     Load the configuration values for the 'common' section.
     """
     try:
-        common_section = CONFIG_TOML[COMMON_SECTION]
+        common_section = CONFIG_TOML.get(COMMON_SECTION, {})
 
         # Check for IMPORT directive (common section in a separate file)
         common_section_import_file = common_section.get(IMPORT, None)
@@ -98,18 +103,18 @@ def load_config_common() -> ConfigCommon:
         ]:
             if args_parser_value is not None:
                 common_section[key_name] = args_parser_value
-                print_log(
-                    f"Using '{key_name}' provided on command line"
-                    f" ('{args_parser_value}')"
-                )
+                print_log(f"Using '{key_name}' provided on command line")
             elif os.environ.get(env_var_name, None) is not None:
                 common_section[key_name] = os.environ[env_var_name]
-                print_log(
-                    f"Using value of '{env_var_name}' environment variable "
-                    f"for '{key_name}'"
-                )
+                print_log(f"Using '{key_name}' provided via the environment")
 
-        url = substitute_variable_str(common_section.get(URL, DEFAULT_URL))
+        # Provide default values for namespace and tag
+        if common_section.get(NAMESPACE, None) is None:
+            common_section[NAMESPACE] = "{{username}}_namespace"
+        if common_section.get(NAME_TAG, None) is None:
+            common_section[NAME_TAG] = "{{username}}_tag"
+
+        url = process_variable_substitutions(common_section.get(URL, DEFAULT_URL))
         if url != DEFAULT_URL:
             print_log(f"Using the YellowDog API at: {url}")
 
@@ -118,13 +123,13 @@ def load_config_common() -> ConfigCommon:
         # substitutions for the items in its dictionary each time it's
         # called
         add_substitutions(subs={URL: url})
-        key = substitute_variable_str(common_section[KEY])
+        key = process_variable_substitutions(common_section[KEY])
         add_substitutions(subs={KEY: key})
-        secret = substitute_variable_str(common_section[SECRET])
+        secret = process_variable_substitutions(common_section[SECRET])
         add_substitutions(subs={SECRET: secret})
-        namespace = substitute_variable_str(common_section[NAMESPACE])
+        namespace = process_variable_substitutions(common_section[NAMESPACE])
         add_substitutions(subs={NAMESPACE: namespace})
-        name_tag = substitute_variable_str(common_section[NAME_TAG])
+        name_tag = process_variable_substitutions(common_section[NAME_TAG])
         add_substitutions(subs={NAME_TAG: name_tag})
 
         return ConfigCommon(
@@ -163,6 +168,12 @@ def load_config_work_requirement() -> Optional[ConfigWorkRequirement]:
         wr_section = CONFIG_TOML[WORK_REQUIREMENT_SECTION]
     except KeyError:
         return ConfigWorkRequirement()
+
+    # Process any new substitutions after the common config
+    # has been processed
+    for _ in range(TOML_VAR_NESTED_DEPTH):
+        process_variable_substitutions_in_dict_insitu(wr_section)
+
     try:
         # Allow WORKER_TAG if WORKER_TAGS is empty
         worker_tags = wr_section.get(WORKER_TAGS, None)
@@ -174,12 +185,12 @@ def load_config_work_requirement() -> Optional[ConfigWorkRequirement]:
         if worker_tags is not None:
             check_list(worker_tags)
             for index, worker_tag in enumerate(worker_tags):
-                worker_tags[index] = substitute_variable_str(worker_tag)
+                worker_tags[index] = process_variable_substitutions(worker_tag)
 
         wr_data_file = wr_section.get(WR_DATA, None)
         if wr_data_file is not None:
             check_str(wr_data_file)
-            wr_data_file = substitute_variable_str(wr_data_file)
+            wr_data_file = process_variable_substitutions(wr_data_file)
             wr_data_file = pathname_relative_to_config_file(
                 CONFIG_FILE_DIR, wr_data_file
             )
@@ -190,7 +201,7 @@ def load_config_work_requirement() -> Optional[ConfigWorkRequirement]:
             if ARGS_PARSER.executable is None
             else ARGS_PARSER.executable
         )
-        executable = substitute_variable_str(executable)
+        executable = process_variable_substitutions(executable)
 
         task_type = (
             wr_section.get(TASK_TYPE, wr_section.get(TASK_TYPE, None))
@@ -199,7 +210,7 @@ def load_config_work_requirement() -> Optional[ConfigWorkRequirement]:
         )
         if task_type is not None:
             check_str(task_type)
-            task_type = substitute_variable_str(task_type)
+            task_type = process_variable_substitutions(task_type)
 
         csv_file = wr_section.get(CSV_FILE, None)
         csv_files = wr_section.get(CSV_FILES, None)
@@ -215,6 +226,12 @@ def load_config_work_requirement() -> Optional[ConfigWorkRequirement]:
             else ARGS_PARSER.task_batch_size
         )
 
+        task_count = (
+            ARGS_PARSER.task_count
+            if ARGS_PARSER.task_count is not None
+            else wr_section.get(TASK_COUNT, 1)
+        )
+
         return ConfigWorkRequirement(
             always_upload=wr_section.get(ALWAYS_UPLOAD, True),
             args=wr_section.get(ARGS, []),
@@ -241,13 +258,14 @@ def load_config_work_requirement() -> Optional[ConfigWorkRequirement]:
             max_workers=wr_section.get(MAX_WORKERS, None),
             min_workers=wr_section.get(MIN_WORKERS, None),
             outputs_optional=wr_section.get(OUTPUTS_OPTIONAL, []),
+            outputs_other=wr_section.get(OUTPUTS_OTHER, []),
             outputs_required=wr_section.get(OUTPUTS_REQUIRED, []),
             priority=wr_section.get(PRIORITY, 0.0),
             providers=wr_section.get(PROVIDERS, None),
             ram=wr_section.get(RAM, None),
             regions=wr_section.get(REGIONS, None),
             task_batch_size=task_batch_size,
-            task_count=wr_section.get(TASK_COUNT, 1),
+            task_count=task_count,
             task_data=wr_section.get(TASK_DATA, None),
             task_data_file=wr_section.get(TASK_DATA_FILE, None),
             task_group_name=wr_section.get(TASK_GROUP_NAME, None),
@@ -279,6 +297,13 @@ def load_config_worker_pool() -> Optional[ConfigWorkerPool]:
     # as a configuration synonym for 'workerPool'. Check for duplicates.
     wp_section = CONFIG_TOML.get(WORKER_POOL_SECTION, {})
     cr_section = CONFIG_TOML.get(COMPUTE_REQUIREMENT_SECTION, {})
+
+    # Process any new substitutions after the common config
+    # has been processed
+    for _ in range(TOML_VAR_NESTED_DEPTH):
+        process_variable_substitutions_in_dict_insitu(wp_section)
+        process_variable_substitutions_in_dict_insitu(cr_section)
+
     duplicate_keys = set(wp_section.keys()).intersection(set(cr_section.keys()))
     if len(duplicate_keys) != 0:
         print_error(
@@ -292,11 +317,11 @@ def load_config_worker_pool() -> Optional[ConfigWorkerPool]:
         return ConfigWorkerPool()
 
     try:
-        worker_tag = substitute_variable_str(wp_section.get(WORKER_TAG, None))
-        worker_pool_data_file = substitute_variable_str(
+        worker_tag = process_variable_substitutions(wp_section.get(WORKER_TAG, None))
+        worker_pool_data_file = process_variable_substitutions(
             wp_section.get(WORKER_POOL_DATA_FILE, None)
         )
-        compute_requirement_data_file = substitute_variable_str(
+        compute_requirement_data_file = process_variable_substitutions(
             wp_section.get(COMPUTE_REQUIREMENT_DATA_FILE, None)
         )
         if (
@@ -322,9 +347,7 @@ def load_config_worker_pool() -> Optional[ConfigWorkerPool]:
                 COMPUTE_REQUIREMENT_BATCH_SIZE, CR_BATCH_SIZE_DEFAULT
             ),
             compute_requirement_data_file=compute_requirement_data_file,
-            idle_node_shutdown_enabled=wp_section.get(IDLE_NODE_SHUTDOWN_ENABLED, True),
             idle_node_shutdown_timeout=wp_section.get(IDLE_NODE_SHUTDOWN_TIMEOUT, 5.0),
-            idle_pool_shutdown_enabled=wp_section.get(IDLE_POOL_SHUTDOWN_ENABLED, True),
             idle_pool_shutdown_timeout=wp_section.get(IDLE_POOL_SHUTDOWN_TIMEOUT, 30.0),
             images_id=wp_section.get(IMAGES_ID, None),
             instance_tags=wp_section.get(INSTANCE_TAGS, None),
@@ -335,7 +358,7 @@ def load_config_worker_pool() -> Optional[ConfigWorkerPool]:
             max_nodes_set=(False if wp_section.get(MAX_NODES) is None else True),
             min_nodes=wp_section.get(MIN_NODES, 0),
             min_nodes_set=(False if wp_section.get(MIN_NODES) is None else True),
-            name=substitute_variable_str(
+            name=process_variable_substitutions(
                 wp_section.get(WP_NAME, None),
             ),
             node_boot_timeout=wp_section.get(NODE_BOOT_TIMEOUT, 10.0),
